import openai
from google.cloud import bigquery
from pydantic import BaseModel

from nlbq.config import get_settings

DEFAULT_MODEL = "gpt-3.5-turbo"

settings = get_settings()


class BytesInfo(BaseModel):
    bytes_used: int
    human_bytes: str
    queries_per_month: int


def get_bytes_info(bytes_used: int) -> BytesInfo:
    """Convert bytes into human readable figure, calculate queries per month"""
    if bytes_used < 1000000:
        human_bytes = f"{bytes_used/1000:.2f}KB"
    else:
        human_bytes = f"{bytes_used/1000000:.2f}MB"
    free_tier_bytes_per_month = 1024**4  # 1TB
    queries_per_month = free_tier_bytes_per_month // bytes_used
    return BytesInfo(
        bytes_used=bytes_used,
        human_bytes=human_bytes,
        queries_per_month=queries_per_month,
    )


def calculate_cost(prompt_tokens: int, completion_tokens: int, model: str) -> float:
    MODEL_COSTS = {
        "gpt-3.5-turbo": {"prompt": 0.002 / 1000, "completion": 0.002 / 1000},
        "gpt-4": {"prompt": 0.03 / 1000, "completion": 0.06 / 1000},
    }
    if model not in MODEL_COSTS:
        raise ValueError(f"Unknown model: {model}")
    costs = MODEL_COSTS[model]
    return (prompt_tokens * costs["prompt"]) + (completion_tokens * costs["completion"])


class NLBQ:
    """Natural language to BigQuery methods"""

    def __init__(self, model: str = DEFAULT_MODEL) -> None:
        self.model = model
        self.prompt_template = self.get_prompt_template()
        self.client = bigquery.Client.from_service_account_json(
            settings.google_application_credentials
        )

    def get_prompt_template(self) -> str:
        """returns the prompt template as a string, with comments removed"""
        lines = open(settings.prompt_template_file).read().split("\n")
        uncommented_lines = [line for line in lines if not line.strip().startswith("#")]
        return "\n".join(uncommented_lines).strip()

    async def text_to_bq(self, question: str) -> tuple:
        """Use an LLM to convert a question into a BigQuery SQL query"""
        prompt_messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": self.prompt_template % question},
        ]
        resp = await openai.ChatCompletion.acreate(
            model=self.model,
            messages=prompt_messages,
            temperature=0,
            api_key=settings.openai_api_key,
        )

        usage = resp["usage"]
        statement = resp["choices"][0]["message"]["content"].strip()
        cost = calculate_cost(
            usage["prompt_tokens"], usage["completion_tokens"], self.model
        )
        return (statement, cost)

    def dry_run(self, query: str) -> BytesInfo:
        """Report on the data this query would use"""
        job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
        query_job = self.client.query(query, job_config=job_config)
        return get_bytes_info(query_job.total_bytes_processed)

    def execute(self, query: str) -> tuple:
        """Execute the query, return the results"""
        bq_query = self.client.query(query)
        results = bq_query.result()
        field_names = [field.name for field in results.schema]
        rows = [row.values() for row in results]
        return field_names, rows

    async def answer(self, question, statement, results) -> str:
        prompt_messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": self.prompt_template % question},
            {"role": "assistant", "content": statement},
            {
                "role": "system",
                "content": "The SQL statement returned the following results:\n\n"
                + results,
            },
            {
                "role": "system",
                "content": f"Using these results, answer the question: {question}",
            },
            {
                "role": "system",
                "content": "Don't say what the query was, just answer the question.",
            },
        ]
        resp = await openai.ChatCompletion.acreate(
            model=self.model,
            messages=prompt_messages,
            temperature=0,
            api_key=settings.openai_api_key,
        )
        return resp["choices"][0]["message"]["content"].strip()



import json
from datetime import date, datetime
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from nlbq.core import NLBQ


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime | date):
            return o.isoformat()
        return super().default(o)


def generate_prompt_from_dataset_table(
    table_in_dataset: str, prompt_file: Path, nlbq: "NLBQ"
):
    """Generate a prompt file from a dataset and table"""
    dataset_id, table_id = table_in_dataset.split(".")

    fields = {
        field.name: str(field)
        for field in nlbq.client.get_table(table_in_dataset).schema
    }

    column_description_glue = ": add column description here\n"
    content = prompt_file.read_text()
    updated_content = (
        content.replace("{{ project_id }}", nlbq.client.project)
        .replace("{{ dataset }}", dataset_id)
        .replace("{{ table_id }}", table_id)
        .replace("{{ schema }}", "\n".join(fields.values()))
        .replace(
            "{{ columns }}",
            column_description_glue.join(fields.keys()) + column_description_glue,
        )
    )

    preview_rows = list(nlbq.client.list_rows(table_in_dataset, max_results=3))
    preview = json.dumps(
        [dict(row) for row in preview_rows], indent=2, cls=CustomJsonEncoder
    )
    updated_content = updated_content.replace("{{ sample_data }}", preview)

    prompt_file.write_text(updated_content)



import shutil
import sys
from asyncio import run
from functools import wraps
from pathlib import Path
from typing import Annotated

import typer
from tabulate import tabulate

from nlbq.api import serve as api_serve
from nlbq.core import DEFAULT_MODEL, NLBQ
from nlbq.helpers import generate_prompt_from_dataset_table


class AsyncTyper(typer.Typer):
    """
    A subclass of Typer that allows async functions to be used as commands
    https://github.com/tiangolo/typer/issues/88#issuecomment-1478432421
    """

    def async_command(self, *args, **kwargs):
        def decorator(async_func):
            @wraps(async_func)
            def sync_func(*_args, **_kwargs):
                return run(async_func(*_args, **_kwargs))

            self.command(*args, **kwargs)(sync_func)
            return async_func

        return decorator


app = AsyncTyper()


@app.async_command()
async def ask(query: str, model: str = DEFAULT_MODEL) -> str:
    """Convert question to BQ query, report on bytes used, offer to execute"""
    nlbq = NLBQ(model=model)
    print(f"Converting your question to a BigQuery query, using {nlbq.model}..." "")
    statement, cost = await nlbq.text_to_bq(query)
    print(f"\n{statement}\n")
    data = nlbq.dry_run(statement)
    print(
        f"This query will process {data.human_bytes}.",
        f"You could run {data.queries_per_month} queries like this per month.",
        f"Building this query cost ${cost:.4f}.",
    )
    if input("Do you want to run it? (Y/n) ").lower() not in ["y", ""]:
        sys.exit(0)
    fields, rows = nlbq.execute(statement)
    print("\n" + tabulate(rows, headers=fields, tablefmt="pipe") + "\n")


def validate_table_format(table_in_dataset: str) -> str:
    """Validate dataset format"""
    if table_in_dataset and "." not in table_in_dataset:
        raise typer.BadParameter(
            "The target table value must be in the format dataset.table_id"
        )
    return table_in_dataset


@app.command()
def init(
    table_in_dataset: Annotated[
        str,
        typer.Option(
            "--table",
            "-t",
            help="The target table in the form of dataset.table_id",
            callback=validate_table_format,
        ),
    ] = None
):
    """Create stub files: prompt.txt, index.html, Dockerfile"""
    package_dir = Path(__file__).parent
    current_dir = Path.cwd()
    prefix_success = typer.style("✔", fg=typer.colors.GREEN, bold=True)
    prefix_error = typer.style("✘", fg=typer.colors.RED, bold=True)

    for file in ["Dockerfile", "index.html"]:
        dest = current_dir / file
        if dest.exists():
            typer.echo(f"{prefix_error} {file} already exists here")
            raise typer.Abort()
        else:
            shutil.copyfile(package_dir / "templates" / file, dest)
            typer.echo(f"{prefix_success} Created {file} at {dest}")

    prompt_file = current_dir / "prompt.txt"
    if prompt_file.exists():
        typer.echo(f"{prefix_error} {file} already exists here")
        raise typer.Abort()

    if table_in_dataset:
        shutil.copyfile(package_dir / "templates" / "prompt_template.txt", prompt_file)
        print(f"⠹ Creating prompt.txt using {table_in_dataset}", end="\r")
        generate_prompt_from_dataset_table(table_in_dataset, prompt_file, NLBQ())
    else:
        shutil.copyfile(package_dir / "templates" / "prompt.txt", prompt_file)

    typer.echo(f"{prefix_success} Created prompt.txt at {prompt_file}")


@app.command()
def serve():
    """Start a FastAPI server"""
    api_serve()


def cli_wrapper():
    """Wrapper for flit entrypoint"""
    app()


if __name__ == "__main__":
    cli_wrapper()



from functools import lru_cache

from pydantic import BaseSettings


class Settings(BaseSettings):
    openai_api_key: str = None
    google_application_credentials: str = None
    uvicorn_host: str = "0.0.0.0"  # noqa: S104
    uvicorn_port: int = 8080
    prompt_template_file: str = "prompt.txt"
    cors_origins: list[str] = ["*"]  # e.g. CORS_ORIGINS=["http://localhost:3000"]
    cors_methods: list[str] = ["*"]
    cors_headers: list[str] = ["*"]
    cors_allow_credentials: bool = False

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"


@lru_cache
def get_settings():
    return Settings()



import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from pydantic import BaseModel
from tabulate import tabulate

from nlbq.config import get_settings
from nlbq.core import NLBQ

settings = get_settings()

app = FastAPI()


app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.cors_origins,
    allow_methods=settings.cors_methods,
    allow_headers=settings.cors_headers,
    allow_credentials=settings.cors_allow_credentials,
)


class DryRunRequest(BaseModel):
    question: str
    model: str


class DryRunResponse(BaseModel):
    statement: str
    data: str
    qpm: int
    gpt_cost: float


class StatementRequest(BaseModel):
    statement: str
    question: str


class StatementResponse(BaseModel):
    text_results: str
    html_results: str


class AnswerRequest(BaseModel):
    question: str
    statement: str
    results: str


@app.post("/api/dry_run")
async def dry_run(request_data: DryRunRequest) -> DryRunResponse:
    """Process a text query and return a SQL statement, results, and explanation."""
    nlbq = NLBQ(model=request_data.model)
    statement, cost = await nlbq.text_to_bq(request_data.question)
    bytes_info = nlbq.dry_run(statement)
    return DryRunResponse(
        statement=statement,
        data=bytes_info.human_bytes,
        qpm=bytes_info.queries_per_month,
        gpt_cost=cost,
    )


@app.post("/api/run_statement")
async def run_statement(request_data: StatementRequest) -> StatementResponse:
    """Execute a SQL statement and return the results in HTML and plain text formats."""
    nlbq = NLBQ()
    fields, rows = nlbq.execute(request_data.statement)
    return StatementResponse(
        text_results=tabulate(rows, headers=fields),
        html_results=tabulate(rows, headers=fields, tablefmt="html"),
    )


@app.post("/api/answer")
async def answer(request_data: AnswerRequest):
    """Explain the results, using the supplied question and results."""
    nlbq = NLBQ()
    resp = await nlbq.answer(
        request_data.question, request_data.statement, request_data.results
    )
    return {"answer": resp}


@app.get("/")
async def serve_index():
    """Serve index.html"""
    return FileResponse("index.html")


def serve():
    uvicorn.run(
        "nlbq.api:app",
        host=settings.uvicorn_host,
        port=settings.uvicorn_port,
        workers=1,
    )
